#include <string.h>
#include "Animation.h"
#include "glob.h"

Animation::Animation( void ) { patchSurfaceList_ = new PatchSurfaceList; }

 void
Animation::insert( PatchSurface* s ) {
    patchSurfaceList_->add( s );
}

#define x(y) fprintf(fd,"%s\n",y); 
 void
Animation::header(
    FILE* fd
) {
    x("# x.rib");
    x("");
    x("##RenderMan RIB-Structure 1.0");
    x("version 3.03");
    x("");
//    x("Format 480 360 1");
    x("Format 450 360 1");
    x("# Format 1024 768 1");
    x("#PixelSamples 2 2");
    x("Option \"searchpath\" \"shader\" [\".:../shaders:&\"]");
    fprintf(fd,"Display \"%s\" \"file\" \"rgba\"\n",tif_file);
    x("Projection \"perspective\"  \"fov\" [18]");
    x("#ShadingRate 1");
    x("");
    x("##CameraOrientation 2 20 5 0 0 1.5 -0.0170699 -0.170699 0.985175");
    x("#These two give a snazzy view:");
    x("ConcatTransform [0.995037 -0.0170699 -0.0980286 0 -0.0995037 -0.170699 -0.980286 0 1.86265e-09 0.985175 -0.17155 0 0 0 0 1]");
    x("Translate -1 -10 -3.0");
    x("#These two give a nearly axial better for debugging:");
    x("#ConcatTransform [0.99904 -0.0040699 -0.02500286 0 -0.0255037 -0.040699 -0.995286 0 0.0 0.995 -0.03155 0 0 0 0 1]");
    x("#Translate 0.0 -10 -1.0");
    x("");
    x("WorldBegin");

    if (caustics) {
        x("Attribute \"caustic\" \"maxpixeldist\" [10]");
        x("Attribute \"light\" \"integer nphotons\" 45000");
        x("LightSource \"caustic\" 7");
    }

    if (caustics || lightradius == 0.0) {
	// This lighting set-up is stolen from Larry Gritz' teapots.rib:
	x("AttributeBegin");
	x("Attribute \"light\" \"shadows\" [\"on\"]");
	x("Rotate -20.0 1 0 0");
	x("LightSource \"spotlight\" 1 \"intensity\" [300] \"from\" [ 2 4.5 20 ] \"to\" [ 0 -1.5 0 ] \"coneangle\" [0.35]");
	x("AttributeEnd");
    } else {

	// This area light is adapted from Larry Gritz' cornell.rib:
	x("AttributeBegin");
	x("Surface \"constant\" ");
	x("Attribute \"light\" \"shadows\" \"on\"");
//	fprintf(fd,"Attribute \"light\" \"nsamples\" [%d]",lightradius > 1.0 ? 1000 : 1);
	fprintf(fd,"Attribute \"light\" \"nsamples\" [%d]",lightradius > 1.0 ? 10 : 1);
	x("Surface \"constant\"");
	x("AreaLightSource \"arealight\" 1 \"intensity\" [250000] \"lightcolor\" [1 1 1]");
	{
	    double s = lightradius; // 's' is for 'size' -- the radius of the area light rectangle
	    double d = 278.0; // 'd' is for 'distance' -- of the light from the origin

	    //  This puts the light directly behind the camera:
	    //fprintf(fd,"Patch \"bilinear\" \"P\" [%g %g %g   %g %g %g   %g %g %g   %g %g %g]\n",s,d,-s, s,d,s, -s,d,-s, -s,d,s);

	    //  This puts the light directly below the object:
	    //fprintf(fd,"Patch \"bilinear\" \"P\" [%g %g %g   %g %g %g   %g %g %g   %g %g %g]\n",s,-s,-d, s,s,-d, -s,-s,-d, -s,s,-d);

	    //  This puts the light directly to the left of the object:
	    //fprintf(fd,"Patch \"bilinear\" \"P\" [%g %g %g   %g %g %g   %g %g %g   %g %g %g]\n",-d,s,-s, -d,s,s, -d,-s,-s, -d,-s,s);

	    // These two put the light directly to the right of the object:
	    // x("Rotate 180.0 0 0 1");
	    // fprintf(fd,"Patch \"bilinear\" \"P\" [%g %g %g   %g %g %g   %g %g %g   %g %g %g]\n",-d,s,-s, -d,s,s, -d,-s,-s, -d,-s,s);

	    // This puts the light directly above the object:
	    x("Rotate 180.0 1 0 0");
	    fprintf(fd,"Patch \"bilinear\" \"P\" [%g %g %g   %g %g %g   %g %g %g   %g %g %g]\n",s,-s,-d, s,s,-d, -s,-s,-d, -s,s,-d);



	    // Instead of the above flat surface,
	    // We should be able to make a hemispherical
	    // sky via something like
	    // x("ReverseOrientation");
	    // x("Sphere 50000 0 50000 360");
	    // here as Larry does, but when I try it the objects go
	    // either flat white or flat black. *BLINK*  Whatever...
	}
	x("AttributeEnd");
    }
    x("Illuminate 1 1");

#ifdef SKY
    // The above is the sun, this is the sky -- bigger, darker, bluer :)
    x("AttributeBegin");
    x("Surface \"constant\" ");
    x("Attribute \"light\" \"shadows\" \"on\"");
    x("Attribute \"light\" \"nsamples\" [32]"); // 256 for serious work 
    x("Surface \"constant\"");
    x("Translate 0 0 0");
    x("AreaLightSource \"arealight\" 2 \"intensity\" [15000] \"lightcolor\" [0.1 0.3 1]");
    {
        double s = 400.0; double d = 378.0;

        // This puts the light directly above the object:
        x("Rotate 180.0 0 1 0");
        fprintf(fd,"Patch \"bilinear\" \"P\" [%g %g %g   %g %g %g   %g %g %g   %g %g %g]\n",s,-s,-d, s,s,-d, -s,-s,-d, -s,s,-d);
    }
    x("AttributeEnd");
    x("Illuminate 2 1");
#endif

    // x("LightSource \"indirect\" 5");
    // This is how you turn on BMRT's indirect lighting support.
    // This is an alternative to its radiosity support.  Radiosity
    // is supposed to be faster on simple scenes but less scalable:
    // The only time I tried this, it segfaulted out on me, though. :( 
    // x("LightSource \"indirect\" 5");
    // These values are adapted from Larry Gritz' tank.indirect.rib.
    // They can be set on an object-by-object basis to control the
    // speed/quality tradeoffs of BMRT's indirect lighting:
    // x("Attribute \"indirect\" \"float maxerror\" [0.05]"); // Default [0.25]
    // x("Attribute \"indirect\" \"float maxpixeldist\" [5]"); // Default [20]
    // x("Attribute \"indirect\" \"float nsamples\" [256]");  // Default [256]



#ifdef OLD
    // This sky is stolen from Larry Gritz' tank.arealight.rib:
    x("AttributeBegin"); 
    x("Attribute \"identifier\" \"name\" [\"sky\"]"); 
    x("ConcatTransform [1 0 0 0 0 1 0 0 0 0 1 0 -1573.14 307.099 -988.299 1]"); 
    x("Attribute \"render\" \"visibility\" [3]"); 
    x("AttributeBegin");
    x("Color [0.388235 0.529412 0.192157]");
    x("Orientation \"rh\"");
    x("ReverseOrientation");
    x("Surface \"constant\"");
    x("Attribute \"light\" \"nsamples\" [64]");
    x("Attribute \"light\" \"shadows\" [\"on\"]");
    x("AreaLightSource \"arealight\" 1 \"lightcolor\" [0.388235 0.529412 0.8] \"intensity\" [10000000000]");
    x("Sphere 50000 0 50000 360");
    x("AttributeEnd");
    x("AttributeEnd");
    x("Illuminate 1 1");
    x("");
#endif


#ifdef OLD
    x("Declare \"casts_shadows\" \"string\"");
    x("Attribute \"render\" \"casts_shadows\" [\"opaque\"]");
    x("LightSource \"ambientlight\" 2 \"lightcolor\" [0.02 0.02 0.02]");
#endif


//  x("LightSource \"ambientlight\" 0 \"intensity\" 0.25 \"lightcolor\" [0.2 0.5 1]");

    if (!nofill) {
	x("Attribute \"light\" \"shadows\" [\"on\"]");
	x("LightSource \"distantlight\" 3 \"intensity\" 0.50  \"from\" [ 1  1 0] \"to\" [0 0 0]  \"lightcolor\" [1 1 1]");
	x("LightSource \"distantlight\" 4 \"intensity\" 0.50  \"from\" [-1  1 0] \"to\" [0 0 0]  \"lightcolor\" [1 1 1]");
	x("");
    }


    if (!nofloor) {
	// This photo-shoot style groundplane/background is stolen from Larry Gritz' tank.unlit.rib:
    #ifdef OLD
	x("AttributeBegin"); 
	x("Attribute \"identifier\" \"name\" [\"ground_plane\"]");
	x("ConcatTransform [1 0 0 0 0 1 0 0 0 0 1 0 -1572.35 939.217 -852.302 1]");
	x("Declare \"visibility\" \"integer\"");
	x("Attribute \"render\" \"visibility\" [3]");
	x("Color [0.690196 0.690196 0.690196]");
	x("Orientation \"lh\"");
	x("Patch \"bilinear\" \"P\" [-53994.6 -57974.4 0 -53994.6 57974.4 0 53994.6 -57974.4 0 53994.6 57974.4 0]");
	x("AttributeEnd");
	x("");
    #endif

	// This wooden floor is stolen from Larry Gritz' teapots.rib:
	x("AttributeBegin");
	x("Attribute \"identifier\" \"name\" [\"Floor\"]");
	x("Attribute \"render\" \"casts_shadows\" [\"none\"]");
	x("Declare \"visibility\" \"integer\"");
	x("Attribute \"render\" \"visibility\" [3]");
    #ifdef CAUSTICS // Never got this to work
	x("Attribute \"radiosity\" \"has_caustics\" 1");
    #endif
	x("Scale 10 10 10");
    #ifndef OLD
	x("Declare \"Kr\" \"float\"");
	x("Surface \"superplank\" \"Kr\" 1");
    #else
	x("Illuminate 6 1");
	x("Color [0.690196 0.690196 0.690196]");
	x("Surface \"clay\"");
    #endif
	x("Patch \"bilinear\" \"P\" [-1 -1 0  1 -1 0  -1 1 0  1 1 0]");
	x("AttributeEnd");
	x("");
    }

    fprintf(fd,"Translate 0 0 1.25\n"); // xxx Find a better way!
    x("");
}

 void
Animation::trailer(
    FILE* fd
) {
    x("");
    x("WorldEnd");
}

 void
Animation::saveState(
    FILE* fd,
    char* title
) {
    PatchSurface* s;
    for (s = patchSurfaceList_->first();
         s;
         s = patchSurfaceList_->next(s)
    ) {
        s->saveState( fd, title );
    }
}

 void
Animation::loadState(
    FILE* fd,
    char* title
) {
    PatchSurface* s;
    for (s = patchSurfaceList_->first();
         s;
         s = patchSurfaceList_->next(s)
    ) {
        s->loadState( fd, title );
    }
}

 void
Animation::body(
    FILE* fd
) {
    PatchSurface* s;
    for (s = patchSurfaceList_->first();
         s;
         s = patchSurfaceList_->next(s)
    ) {
        s->rib( fd );
    }
}

 void
Animation::rib(
    FILE* fd
) {
    this->header(  fd );
    this->body(    fd );
    this->trailer( fd );
}
