#include <assert.h>
#include "BasisMatrix.h"

BasisMatrix * catmull_rom;
BasisMatrix * bspline;
BasisMatrix * bezier;

void
BasisMatrix_init( void ) {
    bezier = new BasisMatrix(
        -1.0,  3.0, -3.0, 1.0, 
         3.0, -6.0,  3.0, 0.0, 
        -3.0,  3.0,  0.0, 0.0, 
         0.0,  0.0, -3.0, 3.0,
         "Basis \"bezier\" 3 \"bezier\" 3"
    );
    bspline = new BasisMatrix(
        -1.0/6.0,  3.0/6.0, -3.0/6.0, 1.0/6.0, 
         3.0/6.0, -6.0/6.0,  3.0/6.0, 0.0/6.0, 
        -3.0/6.0,  0.0/6.0,  3.0/6.0, 0.0/6.0, 
         1.0/6.0,  4.0/6.0,  1.0/6.0, 0.0/6.0,
         "Basis \"b-spline\" 1 \"b-spline\" 1"
    );
    catmull_rom = new BasisMatrix(
       -1.0/2.0,  3.0/2.0, -3.0/2.0,  1.0/2.0, 
        2.0/2.0, -5.0/2.0,  4.0/2.0, -1.0/2.0, 
       -1.0/2.0,  0.0/2.0,  1.0/2.0,  0.0/2.0, 
        0.0/2.0,  2.0/2.0,  0.0/2.0,  0.0/2.0,
        "Basis \"catmull-rom\" 1 \"catmull-rom\" 1"
    );
}

BasisMatrix::BasisMatrix(
        double b00, double b01, double b02, double b03,
        double b10, double b11, double b12, double b13,
        double b20, double b21, double b22, double b23,
        double b30, double b31, double b32, double b33,
        char* text
) {
    m_[0][0] = b00; m_[0][1] = b01; m_[0][2] = b02; m_[0][3] = b03;
    m_[1][0] = b10; m_[1][1] = b11; m_[1][2] = b12; m_[1][3] = b13;
    m_[2][0] = b20; m_[2][1] = b21; m_[2][2] = b22; m_[2][3] = b23;
    m_[3][0] = b30; m_[3][1] = b31; m_[3][2] = b32; m_[3][3] = b33;
    text_ = text;
}

 void
BasisMatrix::rib( FILE* fd ) {
    fprintf( fd, text_ );
}

 void
BasisMatrix::m( int i, int j, double d ) {
    assert (i >= 0 && j >= 0 && i < 4 && j < 4);
    m_[i][j] = d;
}

 double
BasisMatrix::m( int i, int j ) {
    assert (i >= 0 && j >= 0 && i < 4 && j < 4);
    return m_[i][j];
}

 void
BasisMatrix::text( char* text ) {
    text_ = text;
}

 char*
BasisMatrix::text( void ) {
    return text_;
}



