#include <stdio.h>
#include <assert.h>
#include <string.h>
#include "CubicPatch.h"
#include "CubicSpline.h"
#include "AbsolutePoint.h"
#include "nsew.h"

CubicPatch::CubicPatch( void ) {
    c_[0][0]=0; c_[0][1]=0;  c_[0][2]=0; c_[0][3]=0;
    c_[1][0]=0; c_[1][1]=0;  c_[1][2]=0; c_[1][3]=0;
    c_[2][0]=0; c_[2][1]=0;  c_[2][2]=0; c_[2][3]=0;
    c_[3][0]=0; c_[3][1]=0;  c_[3][2]=0; c_[3][3]=0;
    dad_ = 0;
    sib_[0] = 0; sib_[1] = 0; sib_[2] = 0; sib_[3] = 0; sib_[4] = 0; sib_[5] = 0;
    kid_[0] = 0; kid_[1] = 0; kid_[2] = 0; kid_[3] = 0;
    next_ = prev_ = this;
    surface_ = 0;
    minSt_ = St( 0.0, 0.0 );
    maxSt_ = St( 1.0, 1.0 );
}

CubicPatch::CubicPatch(
    BasisMatrix*b,
    Point* p00, Point* p01, Point* p02, Point* p03,
    Point* p10, Point* p11, Point* p12, Point* p13,
    Point* p20, Point* p21, Point* p22, Point* p23,
    Point* p30, Point* p31, Point* p32, Point* p33,
    St minSt,
    St maxSt
  ): m_(b) {
    c_[0][0]=p00; c_[0][1]=p01;  c_[0][2]=p02; c_[0][3]=p03;
    c_[1][0]=p10; c_[1][1]=p11;  c_[1][2]=p12; c_[1][3]=p13;
    c_[2][0]=p20; c_[2][1]=p21;  c_[2][2]=p22; c_[2][3]=p23;
    c_[3][0]=p30; c_[3][1]=p31;  c_[3][2]=p32; c_[3][3]=p33;
    dad_ = 0;
    sib_[0] = 0; sib_[1] = 0; sib_[2] = 0; sib_[3] = 0; sib_[4] = 0; sib_[5] = 0;
    kid_[0] = 0; kid_[1] = 0; kid_[2] = 0; kid_[3] = 0;
    next_ = prev_ = this;
    surface_ = 0;
    minSt_ = minSt;
    maxSt_ = maxSt;
}

 bool
CubicPatch::has( Point* point ) {
    int i, j;
    for     (i = 0; i < 4; ++i) {
        for (j = 0; j < 4; ++j) {
            if (c_[i][j] == point)  return 1;
        }
    }
    return 0;
}

 bool
CubicPatch::isEastNeighbor( CubicPatch* p ) {
    int      count = 0;
    int      i, j;
    for     (i = 1; i < 4; ++i) {
        for (j = 0; j < 4; ++j) {
            if (p->c_[i][j] == c_[i-1][j] ) ++count;
        }
    }
    return count == 12;
}

 bool
CubicPatch::isWestNeighbor( CubicPatch* p ) {
    int      count = 0;
    int      i, j;
    for     (i = 0; i < 3; ++i) {
        for (j = 0; j < 4; ++j) {
            if (p->c_[i][j] == c_[i+1][j] ) ++count;
        }
    }
    return count == 12;
}

 bool
CubicPatch::isSouthNeighbor( CubicPatch* p ) {
    int      count = 0;
    int      i, j;
    for     (i = 0; i < 4; ++i) {
        for (j = 1; j < 4; ++j) {
            if (p->c_[i][j] == c_[i][j-1] ) ++count;
        }
    }
    return count == 12;
}

 bool
CubicPatch::isNorthNeighbor( CubicPatch* p ) {
    int      count = 0;
    int      i, j;
    for     (i = 0; i < 4; ++i) {
        for (j = 0; j < 3; ++j) {
            if (p->c_[i][j] == c_[i][j+1] ) ++count;
        }
    }
    return count == 12;
}

 bool
CubicPatch::isOnSouthCap( void ) {
    return c_[0][1] == c_[1][1] 
        && c_[0][1] == c_[2][1] 
        && c_[0][1] == c_[3][1];
}

 bool
CubicPatch::isOnNorthCap( void ) {
    return c_[0][2] == c_[1][2] 
        && c_[0][2] == c_[2][2] 
        && c_[0][2] == c_[3][2];
}

 bool
CubicPatch::isTwistedNeighbor( CubicPatch* p ) {
    int      count = 0;
    int      i, j;
    for     (i = 0; i < 4; ++i) {
        for (j = 0; j < 4; ++j) {
            if (p->c_[i][j] == c_[(i+2)&3][j] ) ++count;
        }
    }
    return count == 16;
}

 bool
CubicPatch::isTwistedNorthNeighbor( CubicPatch* p ) {
    return this->isOnNorthCap() && this->isTwistedNeighbor( p );
}

 bool
CubicPatch::isTwistedSouthNeighbor( CubicPatch* p ) {
    return this->isOnSouthCap() && this->isTwistedNeighbor( p );
}

// in the current test image configuration,
// NSEW relationships are as expected, and
// the patch names run as follows:
//
//        p l h d
//        o k g c
//        n j f b
//        m i e a
//
// Thus:
//   p l h d   are the North cap,
//   m i e a   are the South cap
//   k   is upper-left of the four most visible patches
//   f   is lower-right of those four
// 

// static CubicPatch*
//find( CubicPatch* p, char* name ) {
//    CubicPatch* q;
//    for (q = p;  q;  ) {
//        if (!strcmp( q->name(), name )) return q;
//        q = q->next();
//        assert( q != p );
//    }
//    return 0; // Only to shut off gcc warning.
//}

// static void
//debugDump2( CubicPatch* p, char* name ) {
//    int i;
//    int j;
//    printf("Dump of %s:\n",name);
//    for     (i = 0; i < 4; ++i) {
//        for (j = 0; j < 4; ++j) {
//            printf("   "); p->controlPoint(i,j)->value().print();
//        }
//        printf("\n");
//    }
//}

// static void
//debugDump( CubicPatch* p ) {
//    CubicPatch* a = find( p, "a" );
//    CubicPatch* i = find( p, "i" );
//    debugDump2( a, "a" );
//    debugDump2( i, "i" );
//}

 void
CubicPatch::checkNeighbors( void ) {

    int i;
    int neighborsFound = 0;
    for (i = 0;  i < 6;  ++i)  if (sib_[i]) ++neighborsFound;
    assert( neighborsFound == 4 );

    {   int inverse[ 6 ];
        inverse[ NORTH ] = SOUTH;
        inverse[ SOUTH ] = NORTH;
        inverse[ EAST  ] = WEST;
        inverse[ WEST  ] = EAST;
        inverse[ TWISTED_NORTH ] = TWISTED_NORTH;
        inverse[ TWISTED_SOUTH ] = TWISTED_SOUTH;

        for (i = 0;  i < 6;  ++i) {
            if (sib_[i]) {
                assert( sib_[i]->neighbor( inverse[i] ) == this);
            }
        }
    }
}



 void
CubicPatch::insert( CubicPatch* p ) {
    // Insert new patch p into existing linklist
    p->next_     = this;
    p->prev_     = prev_;
    prev_->next_ = p;
    prev_        = p;
} 

 void
CubicPatch::remove( void ) {
    // Remove self from linklist
    next_->prev_ = prev_;
    prev_->next_ = next_;
    next_ = this;
    prev_ = this;
}

 CubicPatch*
CubicPatch::next( void ) {
    return next_;
} 

 // Map local patch coords to realspace coords:
 Xyzw
CubicPatch::value( double s, double t ) {
    if (s < 0.0) s = 0.0;  if (s > 1.0) s = 1.0;  
    if (t < 0.0) t = 0.0;  if (t > 1.0) t = 1.0;  

    CubicSpline s0( m_, c_[0][0],  c_[0][1],  c_[0][2],  c_[0][3] );
    CubicSpline s1( m_, c_[1][0],  c_[1][1],  c_[1][2],  c_[1][3] );
    CubicSpline s2( m_, c_[2][0],  c_[2][1],  c_[2][2],  c_[2][3] );
    CubicSpline s3( m_, c_[3][0],  c_[3][1],  c_[3][2],  c_[3][3] );

    AbsolutePoint p0( s0.value( t ) );
    AbsolutePoint p1( s1.value( t ) );
    AbsolutePoint p2( s2.value( t ) );
    AbsolutePoint p3( s3.value( t ) );

    CubicSpline t0( m_, &p0,  &p1,  &p2,  &p3 );

    return t0.value( s );
}

 // Map texture to realspace coords:
 Xyzw
CubicPatch::xyzw( St st ) {

    // Find correct patch in surface:
    if( st.s() > maxSt_.s()) {
        assert( sib_[ EAST  ] );
        return  sib_[ EAST ]->xyzw( st );
    }
    if( st.t() > maxSt_.t()) {
        assert( sib_[ NORTH ] );
        return  sib_[ NORTH ]->xyzw( st );
    }

    assert( st.s() >= minSt_.s() );
    assert( st.t() >= minSt_.t() );

    // Convert from surface-wide texture space
    // coords to our local patch coord system:
    double s = (st.s() - minSt_.s()) / (maxSt_.s() - minSt_.s());
    double t = (st.t() - minSt_.t()) / (maxSt_.t() - minSt_.t());

    return this->value( s, t );
}
 // Map texture to realspace coords:
 Xyzw
CubicPatch::xyzw( Rst rst ) {

    // Find correct patch in surface:
    if( rst.s() > maxSt_.s()) {
        assert( sib_[ EAST  ] );
        return  sib_[ EAST ]->xyzw( rst );
    }
    if( rst.t() > maxSt_.t()) {
        assert( sib_[ NORTH ] );
        return  sib_[ NORTH ]->xyzw( rst );
    }

    assert( rst.s() >= minSt_.s() );
    assert( rst.t() >= minSt_.t() );

    // Convert from surface-wide texture space
    // coords to our local patch coord system:
    double s = (rst.s() - minSt_.s()) / (maxSt_.s() - minSt_.s());
    double t = (rst.t() - minSt_.t()) / (maxSt_.t() - minSt_.t());

    // Convert from patch parametric coords
    // to xyz spatial coords:
    Xyzw here = this->value( s, t );

    // Find normal to surface:
    // (I bet there's some more direct and efficient way...)
    Xyzw ne = this->value( s+0.000001, t+0.000001 );
    Xyzw sw = this->value( s-0.000001, t-0.000001 );
    Xyzw nw = this->value( s-0.000001, t+0.000001 );
    Xyzw se = this->value( s+0.000001, t-0.000001 );
    Xyzw ne_dir = ne-sw;
    Xyzw se_dir = se-nw;
    Xyzw up_dir = (ne_dir ^ se_dir);
    Xyzw nup_dir= up_dir.normalize();

    // Return 'here' displaced normal
    // to the the surface by 'r':
    return here + nup_dir * rst.r();    
}

 void
CubicPatch::rib( FILE* fd ) {
    int m;
    int n;
    fprintf(fd,"\nAttributeBegin\n"); 

    m_->rib( fd );

    fprintf(fd,"Patch \"bicubic\" \"P\" ["); 
    for     (m = 0; m < 4; ++m) { 
        for (n = 0; n < 4; ++n) { 
	    Xyzw x = c_[m][n]->value();
            fprintf(fd," %g %g %g ",x.x(), x.y(), x.z() );
        }
    }
    fprintf(fd,"]\n");
    fprintf(fd,"AttributeEnd\n\n"); 
} 

 void
CubicPatch::surface( PatchSurface* surface ) {
    surface_ = surface;
}

 PatchSurface*
CubicPatch::surface( void ) {
    return surface_;
}

 void
CubicPatch::controlPoint( int i, int j, Point* p ) {
    assert(i >= 0 && j >= 0 && i < 4 && j < 4);
    c_[i][j] = p;
}

 Point* 
CubicPatch::controlPoint( int i, int j ) {
    assert(i >= 0 && j >= 0 && i < 4 && j < 4);
    return c_[i][j];
}

 void
CubicPatch::basisMatrix( BasisMatrix * basisMatrix ) {
    m_ = basisMatrix;
}

 BasisMatrix *
CubicPatch::basisMatrix( void ) {
    return m_;
}

 void
CubicPatch::kid( CubicPatch* kid, int i ) {
    assert( i >= 0 && i < 4);
    kid_[i] = kid; 
}

 void
CubicPatch::sib( CubicPatch* sib, int i ) {
    assert( i >= 0 && i < 4);
    sib_[i] = sib; 
}

 void
CubicPatch::dab( CubicPatch* dad ) {
   dad_ = dad; 
}

 CubicPatch*
CubicPatch::sib( int i ) {
    assert( i >= 0 && i < 4);
    return sib_[i];
}

 CubicPatch*
CubicPatch::kid( int i ) {
    assert( i >= 0 && i < 4);
    return kid_[i];
}

 CubicPatch*
CubicPatch::dad( void ) {
    return dad_;
}

 void
CubicPatch::name( char* name ) {
    assert( strlen(name) +1 < MAX_PATCH_NAME);
    strcpy( name_, name );
}

 char*
CubicPatch::name( void ) {
    return name_;
}

 int
CubicPatch::parity( void ) {
    return parity_;
}

 void
CubicPatch::parity( int i ) {
    parity_ = i & 1;
}

void CubicPatch::minSt( St minSt ) { minSt_ = minSt; }
void CubicPatch::maxSt( St maxSt ) { maxSt_ = maxSt; }

St CubicPatch::minSt( void ) { return minSt_; }
St CubicPatch::maxSt( void ) { return maxSt_; }

 void
CubicPatch::neighbor(
    int direction,
    CubicPatch* neighbor
) {
    assert( direction < 6 && direction >= 0 );
    sib_[direction] = neighbor;
}
 CubicPatch*
CubicPatch::neighbor(
    int direction
) {
    assert( direction < 6 && direction >= 0 );
    return sib_[direction];
}


