#ifndef CUBICPATCH_H
#define CUBICPATCH_H

#include <stdio.h>

#include "BasisMatrix.h"
#include "PatchSurface.h"
#include "Point.h"
#include "Shader.h"
#include "St.h"
#include "SurfaceShader.h"
#include "Xyzw.h"

#define MAX_PATCH_NAME 128 

// This is implements a basic
// cubic-spline patch:
class CubicPatch: public Ribabble {

  public:
    CubicPatch(void);
    CubicPatch(
        BasisMatrix*b,
        Point* p00, Point* p01, Point* p02, Point* p03,
        Point* p10, Point* p11, Point* p12, Point* p13,
        Point* p20, Point* p21, Point* p22, Point* p23,
        Point* p30, Point* p31, Point* p32, Point* p33,
        St minSt,
        St maxSt
    );

    void insert( CubicPatch* p );
    void remove( void );

    void controlPoint( int i, int j, Point* p );
    Point* controlPoint( int i, int j );

    void name( char* name );
    char* name( void );

    void surface( PatchSurface* surface );
    PatchSurface* surface( void );

    void basisMatrix( BasisMatrix * basisMatrix );
    BasisMatrix* basisMatrix( void );

    void minSt( St minSt );    St minSt( void );
    void maxSt( St maxSt );    St maxSt( void );

    void kid( CubicPatch* kid, int i );    CubicPatch* kid( int i );
    void sib( CubicPatch* sib, int i );    CubicPatch* sib( int i );
    void dab( CubicPatch* dad        );    CubicPatch* dad( void  );

    int parity( void ); 
    void parity( int i ); 

    // Map texture to realspace coords:
    Xyzw xyzw( St st );
    Xyzw xyzw( Rst rst );

    // xxx Phase this out:
    Xyzw value( double s, double t );

    CubicPatch* next( void );

    void neighbor( int direction, CubicPatch* neighbor );
    CubicPatch* neighbor( int direction );

    void checkNeighbors( void );

    void rib( FILE* fd );

    bool has( Point* point ); // Is is one of our control points?

  private:
    bool isNorthNeighbor( CubicPatch* p );
    bool isWestNeighbor(  CubicPatch* p );
    bool isEastNeighbor(  CubicPatch* p );
    bool isSouthNeighbor( CubicPatch* p );
    bool isTwistedNeighbor(  CubicPatch* p );
    bool isTwistedNorthNeighbor(  CubicPatch* p );
    bool isTwistedSouthNeighbor(  CubicPatch* p );
    bool isOnNorthCap( void );
    bool isOnSouthCap( void );

    char name_[ MAX_PATCH_NAME ];
    BasisMatrix* m_;
    Point* c_[4][4];
    int         parity_;
    PatchSurface* surface_;
    CubicPatch* prev_;
    CubicPatch* next_;
    CubicPatch* dad_;
    CubicPatch* sib_[6];
    CubicPatch* kid_[4];
    // These two give our local
    // patch of texture coord space:
    St          minSt_; 
    St          maxSt_; 
};

#endif // CUBICPATCH_H
