#include <stdio.h>
#include <assert.h>
#include "CubicSpline.h"

 void
CubicSpline::controlPoint( int i, Point* p ) {
    assert( i >= 0 && i < 4 );
    c_[i] = p;
}

 Point* 
CubicSpline::controlPoint( int i ) {
    assert( i >= 0 && i < 4 );
    return c_[i];
}

 void
CubicSpline::basisMatrix( BasisMatrix * basisMatrix ) {
    m_ = basisMatrix;
}

 BasisMatrix *
CubicSpline::basisMatrix( void ) {
    return m_;
}

Xyzw
CubicSpline::value( double t ) {
    if (t < 0.0) t = 0.0;  
    if (t > 1.0) t = 1.0;  
    double d[4] = { t*t*t, t*t, t, 1.0 };
    double w[4] = { 0.0, 0.0, 0.0, 0.0 };
    Xyzw   c[4] = { c_[0]->value(), c_[1]->value(), c_[2]->value(), c_[3]->value() };
    int i,j;
    for     (i = 0; i < 4; ++i) {
        for (j = 0; j < 4; ++j) {
	    w[i] += d[j] * m_->m_[j][i];
        }
    }

    double x = 0.0;
    double y = 0.0;
    double z = 0.0;
    for     (i = 0; i < 4; ++i) {
        x += w[i] * c[i].x();
        y += w[i] * c[i].y();
        z += w[i] * c[i].z();
    }

    return Xyzw( x, y, z );
}


