#include "Matrix44.h"

 void
Matrix44::identity( void ) {
    int i,j;
    for(i = 0;  i < 4;   i++) {
        for (j = 0;   j < 4;   ++j) {
	  m_[i][j] = 0.0;
        }
    }
    for(i = 0;  i < 4;   i++) {
        m_[i][i] = 1.0;
    }
}

 const Xyzw
Matrix44::xyzw( const Xyzw in ) const {
    /* Apply "mat" to "in" resulting in "out". */
    double xi = in.x();
    double yi = in.y();
    double zi = in.z();
    double wi = in.w();
    double x = xi * m_[0][0] + yi * m_[1][0] + zi * m_[2][0] + wi * m_[3][0];
    double y = xi * m_[0][1] + yi * m_[1][1] + zi * m_[2][1] + wi * m_[3][1];
    double z = xi * m_[0][2] + yi * m_[1][2] + zi * m_[2][2] + wi * m_[3][2];
    double w = xi * m_[0][3] + yi * m_[1][3] + zi * m_[2][3] + wi * m_[3][3];
    if (w != 0.0) {
        if (w != 1.0) {
	    w  = 1.0 / w;
	    x *= w;
	    y *= w;
	    z *= w;
	    w  = 1.0;
        }
    }
    return Xyzw( x, y, z, w );
}

 // Apply the matrix to a point:
 const Xyzw
Matrix44::operator *( Xyzw const b ) const {
   return this->xyzw( b );
 }

 // Multiply two matrices:
 const Matrix44
Matrix44::operator *( Matrix44 const b ) const {
    Matrix44 r; // Result

    register int   i; 
    register int   j; 
  
    /* This routine is probably not a performance bottleneck: */ 
    for (i = 0;   i < 4;   ++i) { 
  
        for (j = 0;   j < 4;   ++j) { 
  
            r.m_[ i ][ j ] = ( 
  
                this->m_[ i ][ 0 ]   *   b.m_[ 0 ][ j ] 		+
                this->m_[ i ][ 1 ]   *   b.m_[ 1 ][ j ] 		+
                this->m_[ i ][ 2 ]   *   b.m_[ 2 ][ j ] 		+
                this->m_[ i ][ 3 ]   *   b.m_[ 3 ][ j ] 
            ); 
        } 
    } 

    return r;
}

 Matrix44 const
Matrix44::scale( double s ) const {
    Matrix44 const m(
        s,   0.0, 0.0, 0.0,
        0.0,   s, 0.0, 0.0,
        0.0, 0.0, s,   0.0,
        0.0, 0.0, 0.0, 1.0
    );
    return *this * m;
}

 Matrix44 const
Matrix44::xRotate( double r ) const {
    double c = cos(r);
    double s = sin(r);
    Matrix44 const m(
        1.0, 0.0, 0.0, 0.0,
        0.0, c,   -s , 0.0,
        0.0, s  ,  c , 0.0,
        0.0, 0.0, 0.0, 1.0
    );
    return *this * m;
}

 Matrix44 const
Matrix44::yRotate( double r ) const {
    double c = cos(r);
    double s = sin(r);
    Matrix44 const m(
        c  , 0.0, s  , 0.0,
        0.0, 1.0, 0.0, 0.0,
        -s , 0.0, c  , 0.0,
        0.0, 0.0, 0.0, 1.0
    );
    return *this * m;
}

 Matrix44 const
Matrix44::zRotate( double r ) const {
    double c = cos(r);
    double s = sin(r);
    Matrix44 const m(
        c  ,-s  , 0.0, 0.0,
        s  , c  , 0.0, 0.0,
        0.0, 0.0, 1.0, 0.0,
        0.0, 0.0, 0.0, 1.0
    );
    return *this * m;
}

 Matrix44 const
Matrix44::translate( double x, double y, double z ) const {
    Matrix44 const m(
        1.0, 0.0, 0.0, 0.0,
        0.0, 1.0, 0.0, 0.0,
        0.0, 0.0, 1.0, 0.0,
        x  , y  , z  , 1.0
    );
    return *this * m;
}

 Matrix44 const
Matrix44::scale( Xyzw about, double s ) const {
    double x = about.x();
    double y = about.y();
    double z = about.z();
    return this->translate(-x,-y,-z).scale(s).translate(x,y,z);
}

 Matrix44 const
Matrix44::xRotate( Xyzw about, double r ) const {
    double x = about.x();
    double y = about.y();
    double z = about.z();
    return this->translate(-x,-y,-z).xRotate(r).translate(x,y,z);
}

 Matrix44 const
Matrix44::yRotate( Xyzw about, double r ) const {
    double x = about.x();
    double y = about.y();
    double z = about.z();
    return this->translate(-x,-y,-z).yRotate(r).translate(x,y,z);
}

 Matrix44 const
Matrix44::zRotate( Xyzw about, double r ) const {
    double x = about.x();
    double y = about.y();
    double z = about.z();
    return this->translate(-x,-y,-z).zRotate(r).translate(x,y,z);
}


