#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include "PatchSurface.h"
#include "glob.h"

PatchSurface::PatchSurface( char* name ) {
    name_ = name;
    shader_ = 0;
    cubicPatchList_ = new CubicPatchList;
    next_ = prev_ = this;
    pointList_ = new PointList();
}

PatchSurface::PatchSurface( char* name, CubicPatch* root ) {
    name_ = name;
    shader_ = 0;
    cubicPatchList_ = new CubicPatchList;
    next_ = prev_ = this;
    cubicPatchList_->add( root );
    pointList_ = new PointList();
}

PatchSurface::PatchSurface( char* name, Shader* shader ) {
    name_ = name;
    shader_ = shader;
    cubicPatchList_ = new CubicPatchList;
    next_ = prev_ = this;
    pointList_ = new PointList();
}

 void
PatchSurface::insert( CubicPatch* p ) {
    cubicPatchList_->add( p );
    p->surface( this );
}

 void
PatchSurface::insert( PatchSurface* p ) {
    // Insert new surface p into existing linklist
    p->next_     = this;
    p->prev_     = prev_;
    prev_->next_ = p;
    prev_        = p;
}

 Xyzw
PatchSurface::xyzw( Rst rst ) {
    assert( cubicPatchList_->length() > 0 );
    return cubicPatchList_->nth(0)->xyzw( rst );
}

 Xyzw
PatchSurface::xyzw( St st ) {
    assert( cubicPatchList_->length() > 0 );
    return cubicPatchList_->nth(0)->xyzw( st );
}

 void
PatchSurface::addPoint( Point* p ) {
    pointList_->add( p );
}

 void
PatchSurface::checkNeighbors( void ) {
    CubicPatch* p;
    for (p = cubicPatchList_->first();
         p;
         p = cubicPatchList_->next(p)
    ){
        p->checkNeighbors();
    }
}

 void
PatchSurface::saveState(
    FILE* fd,
    char* title
) {
    char buf[ 1024 ];
    if (!title || !*title) {
        strcpy( buf, name_ );
    } else {
        sprintf( buf, "%s.%s", title, name_ );
    }
    assert( strlen( buf ) < 1024 );

    fprintf(fd,"PatchSurface %s pts=%d\n",buf,pointList_->length());
    Point* p;
    int i = 0;
    for(p = pointList_->first();
        p;
        p = pointList_->next(p), ++i
    ) {
        char buf2[ 1024 ];
        sprintf(buf2,"%s.%d", buf, i );
        assert( strlen( buf2 ) < 1024 );
        p->saveState( fd, buf2 );
    }
}

 void
PatchSurface::loadState(
    FILE* fd,
    char* title
) {
    char input[ 1024 ];
    fgets( input, 1024, fd );
    assert(strlen(input)<1023);

    char buf[ 1024 ];
    if (!title || !*title) {
        strcpy( buf, name_ );
    } else {
        sprintf( buf, "%s.%s", title, name_ );
    }
    assert( strlen( buf ) < 1024 );

    char buf2[ 1024 ];
    sprintf(buf2,"PatchSurface %s pts=",buf);
    int len = strlen( buf2 );
    assert( len < 1024 );
    assert( !strncmp( input, buf2, len ) );
    int pts = atoi( input+len );
    assert( pts == pointList_->length() );
    Point* p;
    int i = 0;
    for(p = pointList_->first();
        p;
        p = pointList_->next(p), ++i
    ) {
        char buf2[ 1024 ];
        sprintf(buf2,"%s.%d", buf, i );
        assert( strlen( buf2 ) < 1024 );
        p->loadState( fd, buf2 );
    }
}

 void
PatchSurface::rib( FILE* fd ) {
    CubicPatch* p;

    fprintf(fd,"AttributeBegin\n"); 

    if (caustics) {
        fprintf(fd,"Declare \"specularcolor\" \"color\"\n");
        fprintf(fd,"Declare \"refractioncolor\" \"color\"\n");
        fprintf(fd,"Declare \"refractionindex\" \"float\"\n");
        fprintf(fd,"Attribute \"caustic\" \"specularcolor\" [1. 1. 1.]\n");
        fprintf(fd,"Attribute \"caustic\" \"refractioncolor\" [1. 1. 1.]\n");
        fprintf(fd,"Attribute \"caustic\" \"refractionindex\" 2.2\n");
    }


    for (p = cubicPatchList_->first();
         p;
         p = cubicPatchList_->next(p)
    ){

        fprintf(fd,"AttributeBegin\n"); 
        if (shader_) {
            shader_->parity( p->parity() );
            shader_->rib( fd );
        }

        p->rib( fd );
        fprintf(fd,"AttributeEnd\n"); 
    }
    fprintf(fd,"AttributeEnd\n"); 
}

 PatchSurface*
PatchSurface::next( void ) {
    return next_;
}

 Shader *
PatchSurface::shader( void ) {
    return shader_;
}

 void
PatchSurface::shader( Shader* shader ) {
    shader_ = shader;
}
    
 char *
PatchSurface::name( void ) {
    return name_;
}

 void
PatchSurface::name( char* name ) {
    int len = strlen( name );
    char* name_ = (char*)malloc( len + 1 );
    strcpy( name_, name );
}
    
