#ifndef PATCHSURFACE_H
#define PATCHSURFACE_H

#include <stdio.h>
#include "Shader.h"
#include "Surface.h"
class PatchSurface;
class CubicPatch;
class CubicPatchList;
#include "CubicPatch.h"
#include "CubicPatchList.h"
#include "PointList.h"

// A surface made out of CubicPatches:
class PatchSurface: public Surface {

  public:
    PatchSurface( char*name, void );
    PatchSurface( char*name, Shader* shader );
    PatchSurface( char*name, CubicPatch* root );

    CubicPatch * root( void );
    void root( CubicPatch* root );
    
    Shader * shader( void );
    void shader( Shader* shader );
    
    void insert( PatchSurface* s );
    void insert( CubicPatch* p );

    void addPoint( Point* p );

    void rib( FILE* fd );

    void checkNeighbors( void );

    PatchSurface* next( void );

    void saveState( FILE* fd, char* title );
    void loadState( FILE* fd, char* title );

    char* name( void );
    void name( char* );

    virtual Xyzw xyzw( St st );
    virtual Xyzw xyzw( Rst rst );

  private:
    // next_/prev_ allow us to be linked together
    // with other PatchSurfaces for the benefit
    // of higher level code.  We should write
    // a collection class of some sort for this
    // purpose and then eliminate these two:
    PatchSurface * next_;
    PatchSurface * prev_;

    // A list of all cubic patches in surface:
    CubicPatchList * cubicPatchList_;

    // shader_ is a simple hook for remembering
    // what RenderMan surface type we should use
    // when generating a .rib representing us:
    Shader     * shader_;

    PointList  * pointList_; 

    char* name_;
};

#endif // SHAPE_H
