#include <stdio.h>
#include <math.h>
#include "VaryWithSymmetry.h"
#include "Hash.h"

VaryWithSymmetry::VaryWithSymmetry(

    double scale,
    double salt,

    bool   mirrorS,
    bool   mirrorT,
    bool   mirrorX,
    bool   mirrorY,
    bool   mirrorZ,

    bool   ignoreS,
    bool   ignoreT,
    bool   ignoreX,
    bool   ignoreY,
    bool   ignoreZ
) {
    scale_        = scale;
    salt_         = salt;

    mirrorS_      = mirrorS;
    mirrorT_      = mirrorT;
    mirrorX_      = mirrorX;
    mirrorY_      = mirrorY;
    mirrorZ_      = mirrorZ;

    ignoreS_      = ignoreS;
    ignoreT_      = ignoreT;
    ignoreX_      = ignoreX;
    ignoreY_      = ignoreY;
    ignoreZ_      = ignoreZ;
}

 static double
drop_noise_bits(
    double d
) {
    if (d > 0)  return   floor( 10000.0 *  d );
    else        return  -floor( 10000.0 * -d );
}

 Rst
VaryWithSymmetry::rst(
    Rst rst,
    Surface* surface
) {
    double r = rst.r();
    double s = rst.s();
    double t = rst.t();

    Xyzw p = surface->xyzw( rst );

    double hs = s    ;
    double ht = t    ;
    double hx = p.x();
    double hy = p.y();
    double hz = p.z();

    if (ignoreS_)  hs = 0.0;
    if (ignoreT_)  ht = 0.0;
    if (ignoreX_)  hx = 0.0;
    if (ignoreY_)  hy = 0.0;
    if (ignoreZ_)  hz = 0.0;

    if (mirrorS_)  hs = fabs(hs-0.5)+0.5;
    if (mirrorT_)  ht = fabs(ht-0.5)+0.5;
    if (mirrorX_)  hx = fabs(hx);
    if (mirrorY_)  hy = fabs(hy);
    if (mirrorZ_)  hz = fabs(hz);

    hs = drop_noise_bits( hs );
    ht = drop_noise_bits( ht );
    hx = drop_noise_bits( hx );
    hy = drop_noise_bits( hy );
    hz = drop_noise_bits( hz );

    r += scale_* 2.0 * (hash->to_0_1(hs,ht,hx,hy,hz,salt_) - 0.5);

    return Rst( r, s, t );
}

